import tensorflow as tf
import tensorflow_models as tfm
from transact.transact_config import TransActConfig


class TensorflowTransAct(tf.keras.Model):
    def __init__(self, transact_config: TransActConfig, training=False):
        super().__init__()
        self.transact_config = transact_config
        self.training = training
        self.action_vocab: list = self.transact_config.action_vocab

        if self.transact_config.concat_candidate_emb:
            transformer_in_dim = (
                self.transact_config.action_emb_dim
                + self.transact_config.item_emb_dim * 2
            )
        else:
            transformer_in_dim = (
                self.transact_config.action_emb_dim + self.transact_config.item_emb_dim
            )

        self.action_type_lookup = tf.keras.layers.IntegerLookup(
            num_oov_indices=1,
            mask_token=-1,
            vocabulary=list(self.transact_config.action_vocab),
            output_mode="int",
        )
        self.action_emb_module = tf.keras.layers.Embedding(
            input_dim=len(self.action_vocab) + 1 + 1,  # oov
            output_dim=self.transact_config.action_emb_dim,
            input_length=self.transact_config.seq_len,
        )

        self.transformer_encoder = tfm.nlp.models.TransformerEncoder(
            num_layers=self.transact_config.num_layer,
            num_attention_heads=self.transact_config.nhead,
            intermediate_size=self.transact_config.dim_feedforward,
            activation="relu",
            dropout_rate=0.0,
            attention_dropout_rate=0.0,
            use_bias=False,
            norm_first=True,
            norm_epsilon=1e-06,
            intermediate_dropout=0.0,
        )

        if self.transact_config.concat_max_pool:
            self.max_pool = tf.keras.layers.GlobalMaxPooling1D()
            self.out_linear = tf.keras.layers.Dense(transformer_in_dim)

    def call(
        self,
        action_type_seq: tf.Tensor,
        item_embedding_seq: tf.Tensor,
        action_time_seq: tf.Tensor,
        request_time: tf.Tensor,
        item_embedding: tf.Tensor,
    ):
        # step 1: get the latest N actions from sequence features
        action_type_seq = action_type_seq[:, : self.transact_config.seq_len]
        item_embedding_seq = item_embedding_seq[:, : self.transact_config.seq_len, :]
        action_time_seq = action_time_seq[:, : self.transact_config.seq_len]

        # step 2: get action embedding
        action_type_seq = self.action_type_lookup(action_type_seq)
        action_emb_tensor = self.action_emb_module(action_type_seq)

        # step 3: create mask that tells transformer which position to be ignored by the attention.
        key_padding_mask = tf.math.less_equal(action_type_seq, 0)

        # mask actions that happened in a time window before the request time
        request_time = tf.expand_dims(request_time, axis=-1)
        request_time = tf.broadcast_to(request_time, tf.shape(action_time_seq))

        # randomly sample a time window to introduce randomness
        rand_time_window_ms = tf.random.uniform(
            shape=(),
            minval=0,
            maxval=self.transact_config.time_window_ms,
            dtype=tf.int32,
        )
        short_time_window_idx_trn = tf.math.less(
            request_time - action_time_seq, rand_time_window_ms
        )
        short_time_window_idx_eval = tf.math.less(request_time - action_time_seq, 0)

        # adjust the mask accordingly
        if self.training:
            key_padding_mask = self._adjust_mask(
                key_padding_mask, short_time_window_idx_trn
            )
        else:
            key_padding_mask = self._adjust_mask(
                key_padding_mask, short_time_window_idx_eval
            )

        # step 4: concat seq embedding with action embedding and candidate embedding
        action_pin_emb = tf.concat((action_emb_tensor, item_embedding_seq), axis=2)

        if self.transact_config.concat_candidate_emb:
            item_embedding_expanded = tf.expand_dims(item_embedding, axis=1)
            item_embedding_expanded = tf.tile(
                item_embedding_expanded, multiples=[1, self.seq_len, 1]
            )
            action_pin_emb = tf.concat(
                (action_pin_emb, item_embedding_expanded), axis=-1
            )

        key_padding_mask = tf.expand_dims(key_padding_mask, axis=-1)
        tfmr_out = self.transformer_encoder(
            encoder_inputs=action_pin_emb, attention_mask=key_padding_mask
        )

        # output_concat = []
        # if self.transact_config.concat_max_pool:
        # Apply max pooling to the transformer output
        pooled_max = self.max_pool(tfmr_out)
        pooled_out = self.out_linear(pooled_max)
        # output_concat.append(pooled_out)

        # if self.transact_config.latest_n_emb > 0:
        #     tfmr_out = tfmr_out[:, :self.transact_config.latest_n_emb]
        # output_concat.append(tf.reshape(tfmr_out, [tf.shape(tfmr_out)[0], -1]))
        # output = tf.keras.layers.Concatenate(axis=1)(output_concat)
        # return output
        return pooled_out

    def _adjust_mask(self, mask, short_time_window_idx):
        # Make sure not all actions in the sequence are masked
        mask = tf.cast(mask, tf.int64)
        short_time_window_idx = tf.cast(short_time_window_idx, tf.int64)

        mask = tf.bitwise.bitwise_or(mask, short_time_window_idx)

        # Set the first column of the mask to zeros
        mask = tf.concat([tf.zeros_like(mask[:, :1]), mask[:, 1:]], axis=1)

        # Create a new attention mask with -inf for masked positions
        new_attn_mask = tf.where(tf.cast(mask, tf.bool), float(-1), 0.0)

        return new_attn_mask



from dataclasses import dataclass


@dataclass
class TransActConfig:
    """
    Configuration class to build a TransAct PyTorch module.

    :param seq_len: Length of the input sequence
    :param time_window_ms: Time window in milliseconds for random window mask
    :param latest_n_emb: Number of latest embeddings to use in output
    :param concat_candidate_emb: Whether to concatenate candidate embeddings with user sequence
    :param concat_max_pool: Whether to apply max pooling to the output of the transformer encoder and append it to output
    :param action_vocab: Vocabulary of user actions
    :param action_emb_dim: Dimension of user action embeddings
    :param item_emb_dim: Dimension of item embeddings
    :param num_layer: Number of TransformerEncoderLayer
    :param nhead: Number of heads in the TransformerEncoderLayer
    :param dim_feedforward: Feed forward dimension of the TransformerEncoderLayer
    """

    action_vocab: list = range(0, 20)
    action_emb_dim: int = 32
    item_emb_dim: int = 32
    time_window_ms: int = 1000 * 60 * 60 * 1
    latest_n_emb: int = 10
    seq_len: int = 5
    concat_candidate_emb: bool = False
    concat_max_pool: bool = True
    num_layer: int = 2
    nhead: int = 2
    dim_feedforward: int = 32



import os.path

from setuptools import find_packages, setup


def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="transact-tf",
    version=get_version("transact/version.py"),
    description="An implementation of TransAct: Transformer-based Realtime User Action Model for Recommendation",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    url="https://github.com/awadalaa/TransAct",
    author="Alaa Awad",
    author_email="alaa.awad.mail@gmail.com",
    install_requires=["tensorflow>=2.9", "tf-models-official==2.9"],
    extras_require={
        "dev": [
            "check-manifest",
            "twine",
            "numpy",
            "black",
            "pytest",
            "typing-extensions",
        ],
    },
)
